/*
Copyright (c) 2022 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#if !defined(_WIN32) && !defined(__CYGWIN__)
#  define _POSIX_C_SOURCE 200809L // for declaration of `strnlen()` on POSIX-compliant implementations
#endif
#include "ccsv.h"
#include <stdint.h>
#include <stdio.h>
#include <string.h>

// clang-format off
#if defined(_WIN32) || defined(__CYGWIN__) // Windows
#  include <share.h>
#  define CSV_CHECK_MAX_ (8 * 1024) // maximum number of bytes in the raw data to search for a null byte
#  define csv_fopen_(_filename, _mode)           _fsopen(_filename, _mode, _SH_DENYNO)
#  define csv_wfopen_(_filename, _mode)          _wfsopen(_filename, _mode, _SH_DENYNO)
#  define csv_fseek_(_stream, _offset, _origin)  _fseeki64(_stream, _offset, _origin)
#  define csv_ftell_(_stream)                    _ftelli64(_stream)
typedef long long csv_long_t;
#else // POSIX
#  define CSV_CHECK_MAX_ (sizeof(wchar_t)) // 4 bytes to search for a null byte is enough, any `wchar_t` value contains at least one null byte because the highest Unicode code point is 0x0010FFFF
#  define csv_fopen_(_filename, _mode)           fopen(_filename, _mode)
#  define csv_wfopen_(_filename, _mode)          NULL
#  define csv_fseek_(_stream, _offset, _origin)  fseek(_stream, _offset, _origin)
#  define csv_ftell_(_stream)                    ftell(_stream)
typedef long csv_long_t;
#endif
// clang-format on

#define CSV_RECORDS_CHUNK_ 128 // Initially allocated number of `csv_record_t` pointers, increased by half of the former value in each reallocation.
#define CSV_FIELDS_CHUNK_ 32 // In the first record, number of initially allocated `csv_field_t` pointers, increased by this value in each reallocation. In subsequent records, the `maxFields` value is initially used.

#define CSV_IN_RECORD_ 0x1
#define CSV_IN_FIELD_ 0x2
#define CSV_IN_QUOTES_ 0x4
#define CSV_PREV_QUOTE_ 0x8
#define CSV_SKIP_COMMENT_ 0x10

#define flags_clear_(_flags) (_flags = 0)
#define flags_set_(_flags, _flag) (_flags |= (_flag))
#define flags_del_(_flags, _flag) (_flags &= ~(_flag))
#define flags_flip_(_flags, _flag) (_flags ^= (_flag))
#define flags_get_(_flags, _flag) (_flags & (_flag))

#define assign_increment_(_isNarrow, _pTxtIter, _chr) \
  ((_isNarrow) ?                                      \
     (*(*(char **)_pTxtIter)++ = (char)(_chr)) :      \
     (*(*(wchar_t **)_pTxtIter)++ = (wchar_t)(_chr)))

#if defined(__GNUC__) // e.g. gcc, clang
#  define csv_want_inline_ inline __attribute__((always_inline))
#elif defined(_MSC_VER) // MSVC
#  define csv_want_inline_ __forceinline
#else
#  define csv_want_inline_ inline
#endif

#if defined(__GNUC__)
#  pragma GCC diagnostic push
#  pragma GCC diagnostic ignored "-Wdeclaration-after-statement" // mixing declarations and code is incompatible with standards before C99, however we require C11 anyway
#  if defined(__clang__)
#    pragma GCC diagnostic ignored "-Wc99-extensions" // compound literals are a C99-specific feature
#  endif
#elif defined(_MSC_VER)
#  pragma warning(push)
#  pragma warning(disable : 4711 5045 6001 6308 6386 6387) // function selected for inlining; compiler may add Spectre mitigation; use of uninitialized memory; realloc may return NULL; possible buffer overrun; parameter could be NULL
#endif

// ***** private functions: *****

static csv_raw_t *alloc_raw_(const size_t dataSize)
{
  csv_raw_t *pRaw = (csv_raw_t *)malloc(sizeof(size_t) + dataSize + 2 * sizeof(wchar_t)); // this block of memory is reused for the `txt` or `wtxt` members of all `csv_field_t` objects
  if (!pRaw)
    abort();

  pRaw->dataSize = dataSize;
  wchar_t *const tail = (wchar_t *)(void *)(pRaw->data + dataSize);
  *tail = L'\0';
  tail[1] = L'\0';
  return pRaw;
}

static csv_raw_t *read_csv_raw_(const bool narrowName, const void *const fileName)
{
  if (!fileName)
    return NULL;

  FILE *const stream = narrowName ? csv_fopen_((const char *)fileName, "rb") : csv_wfopen_((const wchar_t *)fileName, L"rb");
  if (!stream)
    return NULL;

  csv_long_t size = 0;
  if (csv_fseek_(stream, 0, SEEK_END) != 0 || (size = csv_ftell_(stream)) < 0 || csv_fseek_(stream, 0, SEEK_SET) != 0)
  {
    fclose(stream);
    return NULL;
  }

  csv_raw_t *pRaw = alloc_raw_((size_t)size);
  if (fread(pRaw->data, 1, pRaw->dataSize, stream) != pRaw->dataSize)
  {
    free(pRaw);
    fclose(stream);
    return NULL;
  }

  fclose(stream);
  return pRaw;
}

static size_t get_bom_len_(const csv_raw_t *const pRaw)
{
  if (pRaw->dataSize >= sizeof(wchar_t) && *(const wchar_t *)(const void *)(pRaw->data) == L'\xFEFF')
    return sizeof(wchar_t);

  const char *const pBegin = (const char *)pRaw->data;
  if (pRaw->dataSize >= 3 && *pBegin == '\xEF' && pBegin[1] == '\xBB' && pBegin[2] == '\xBF')
    return 3;

  return 0;
}

static bool get_is_narrow_(const csv_raw_t *const pRaw, const size_t bomLen, const bool fallbackNarrow)
{
  if (bomLen)
    return (bomLen == 3); // `true` if UTF-8 BOM found, `false` if UTF-16/UTF-32 BOM found

  if (pRaw->dataSize % sizeof(wchar_t)) // Whenever data is not `wchar_t`-aligned, it is `char` data. This simple test reduces the probability of subsequently running `strnlen` by 50% (Windows) or even 75% (*nix).
    return true;

  if (pRaw->dataSize < CSV_CHECK_MAX_ ? //                                       Whenever data consists of `wchar_t`, null-bytes are found in at least every ASCII character (on *nix even in all characters) ...
        strnlen((const char *)pRaw->data, pRaw->dataSize) != pRaw->dataSize : // ... and `strnlen` would stop early. The risk that CSV data does not contain ASCII characters is negligible low, because ...
        strnlen((const char *)pRaw->data, CSV_CHECK_MAX_) != CSV_CHECK_MAX_) //  ... even '\n' is ASCII (besides of digits, decimal points, quotation marks, spaces, and likely also the used separator).
    return false;

  return fallbackNarrow;
}

static int comp_narrow_names_(const void *a, const void *b)
{
  return strcmp((const char *)(((const csv_col_head_t *)a)->colName), (const char *)(((const csv_col_head_t *)b)->colName));
}

static int comp_wide_names_(const void *a, const void *b)
{
  return wcscmp((const wchar_t *)(((const csv_col_head_t *)a)->colName), (const wchar_t *)(((const csv_col_head_t *)b)->colName));
}

static void create_sorted_header_(csv_t *pCsv)
{
  pCsv->colHeads = (csv_col_head_t *)malloc(pCsv->records->fieldsCount * sizeof(csv_col_head_t));
  if (!pCsv->colHeads)
    abort();

  for (size_t idx = 0; idx != pCsv->records->fieldsCount; ++idx)
  {
    pCsv->colHeads[idx].colName = pCsv->records->fields[idx].txt;
    pCsv->colHeads[idx].colIdx = idx;
  }

  qsort(pCsv->colHeads, pCsv->records->fieldsCount, sizeof(csv_col_head_t), pCsv->isNarrow ? comp_narrow_names_ : comp_wide_names_);
}

static void check_increase_records_(csv_t *const pCsv, size_t *const pRecordsCapa, size_t *const pRecordsChunk)
{
  if (pCsv->recordsCount + 1 < *pRecordsCapa)
    return;

  pCsv->records = (csv_record_t *)realloc(pCsv->records, *pRecordsChunk * sizeof(csv_record_t));
  if (!pCsv->records)
    abort();

  *pRecordsCapa = *pRecordsChunk;
  *pRecordsChunk += (*pRecordsChunk >> 1) + 1;
}

static void check_increase_fields_(csv_record_t *const pRecord, size_t *const pFieldsCapa, const size_t fildsChunk)
{
  if (pRecord->fieldsCount + 1 < *pFieldsCapa)
    return;

  *pFieldsCapa += fildsChunk;
  pRecord->fields = (csv_field_t *)realloc(pRecord->fields, *pFieldsCapa * sizeof(csv_field_t));
  if (!pRecord->fields)
    abort();
}

static void check_update_minmax_fields_(csv_t *const pCsv, const csv_record_t *const pRecord)
{
  if (!pRecord)
    return;

  if (pRecord->fieldsCount < pCsv->minFields || !pCsv->minFields)
    pCsv->minFields = pRecord->fieldsCount;

  if (pRecord->fieldsCount > pCsv->maxFields)
    pCsv->maxFields = pRecord->fieldsCount;
}

static void update_record_dat_(csv_t *const pCsv, csv_record_t **const ppRecord)
{
  *ppRecord = pCsv->records + pCsv->recordsCount;
  (*ppRecord)->fields = NULL;
  (*ppRecord)->fieldsCount = 0;
  ++pCsv->recordsCount;
}

static void update_field_dat_(csv_record_t *const pRecord, csv_field_t **const ppField, void *const pTxt)
{
  *ppField = pRecord->fields + pRecord->fieldsCount;
  (*ppField)->txt = (char *)pTxt;
  (*ppField)->txtLength = 0;
  ++pRecord->fieldsCount;
}

csv_want_inline_ static void handle_single_char_(const bool isNarrow, int *const pFlags, const int chr, const csv_param_t *const pParam, void **const pTxtIter, size_t *const pTxtCnt, csv_t *const pCsv, csv_record_t **const ppRecord, size_t *const pRecordsCapa, size_t *const pRecordsChunk, csv_field_t **const ppField, size_t *const pFieldsCapa, size_t *const pFieldsChunk)
{
  if (!flags_get_(*pFlags, CSV_IN_RECORD_))
  {
    if (flags_get_(*pFlags, CSV_SKIP_COMMENT_))
    {
      if (chr == '\n')
        flags_clear_(*pFlags);

      return;
    }

    if (chr == pParam->commentPrefix)
    {
      flags_set_(*pFlags, CSV_SKIP_COMMENT_);
      return;
    }

    if (pParam->skipEmpty && (chr == '\n' || chr == '\r'))
      return;

    check_update_minmax_fields_(pCsv, *ppRecord);
    check_increase_records_(pCsv, pRecordsCapa, pRecordsChunk);
    update_record_dat_(pCsv, ppRecord);
    flags_set_(*pFlags, CSV_IN_RECORD_);
    *pFieldsCapa = 0;
    if (pCsv->maxFields > 1)
      *pFieldsChunk = pCsv->maxFields;
  }

  if (!flags_get_(*pFlags, CSV_IN_FIELD_))
  {
    check_increase_fields_(*ppRecord, pFieldsCapa, *pFieldsChunk);
    update_field_dat_(*ppRecord, ppField, *pTxtIter);
    flags_set_(*pFlags, CSV_IN_FIELD_);
  }

  switch (chr)
  {
  case '\n':
    if (flags_get_(*pFlags, CSV_IN_QUOTES_)) // LF in quoted string (belongs to the value of a field)
    {
      assign_increment_(isNarrow, pTxtIter, '\n');
      ++*pTxtCnt;
      flags_del_(*pFlags, CSV_PREV_QUOTE_);
      break;
    }

    assign_increment_(isNarrow, pTxtIter, '\0');
    (*ppField)->txtLength = *pTxtCnt;
    *pTxtCnt = 0;
    flags_clear_(*pFlags);
    break;

  case '\r':
    if (flags_get_(*pFlags, CSV_IN_QUOTES_)) // CR in quoted string (belongs to the value of a field)
    {
      assign_increment_(isNarrow, pTxtIter, '\r');
      ++*pTxtCnt;
    }

    flags_del_(*pFlags, CSV_PREV_QUOTE_);
    break;

  case '"':
    if (!flags_get_(*pFlags, CSV_IN_QUOTES_) && flags_get_(*pFlags, CSV_PREV_QUOTE_)) // treated as second of two consecutive quotes in quoted string (`!flags_get_()` might be confusing but is no mistake)
    {
      assign_increment_(isNarrow, pTxtIter, '"');
      ++*pTxtCnt;
    }

    flags_flip_(*pFlags, CSV_IN_QUOTES_);
    flags_set_(*pFlags, CSV_PREV_QUOTE_);
    break;

  default:
    if (chr == pParam->separator && !flags_get_(*pFlags, CSV_IN_QUOTES_)) // separator in unquoted string
    {
      assign_increment_(isNarrow, pTxtIter, '\0');
      (*ppField)->txtLength = *pTxtCnt;
      *pTxtCnt = 0;
      flags_del_(*pFlags, CSV_IN_FIELD_ | CSV_PREV_QUOTE_);
      break;
    }

    assign_increment_(isNarrow, pTxtIter, chr);
    ++*pTxtCnt;
    flags_del_(*pFlags, CSV_PREV_QUOTE_);
  }
}

static void handle_end_of_data_(const int flags, void **const pTxtIter, csv_t *const pCsv, csv_record_t *const pRecord, csv_field_t **const ppField, size_t *const pFieldsCapa, const size_t txtCnt)
{
  if (flags_get_(flags, CSV_IN_FIELD_)) // no new line and no separator at the end of the raw data
    (*ppField)->txtLength = txtCnt;
  else if (flags_get_(flags, CSV_IN_RECORD_)) // no new line but a separator at the end of the raw data
  {
    check_increase_fields_(pRecord, pFieldsCapa, 1);
    update_field_dat_(pRecord, ppField, *pTxtIter);
  }

  check_update_minmax_fields_(pCsv, pRecord);
  wchar_t *const tail = *(wchar_t **)pTxtIter;
  *tail = L'\0';
  tail[1] = L'\0';
}

static void parse_narrow_csv_(void *txtIter, const unsigned char *rawIter, const size_t rawSize, int *const pFlags, const csv_param_t *const pParam, csv_t *const pCsv, csv_record_t **const ppRecord, size_t *const pRecordsCapa, size_t *const pRecordsChunk, csv_field_t **const ppField, size_t *const pFieldsCapa, size_t *const pFieldsChunk, size_t *const pTxtCnt)
{
  static const bool isNarrow = true; // The result of `get_is_narrow_` is not passed to this function as it is determined at runtime. Instead, this literal value is used to emphasize that the compiler shall remove any conditional use of `isNarrow` in this branch. (refer to `assign_increment_` which is executed for almost each character)
  pCsv->isNarrow = isNarrow;
  const csv_param_t param = (csv_param_t){ (int)(unsigned char)pParam->separator, (int)(unsigned char)pParam->commentPrefix, pParam->skipEmpty, pParam->fallbackNarrow };
  for (const unsigned char *const rawEnd = rawIter + rawSize; rawIter != rawEnd; ++rawIter)
    handle_single_char_(isNarrow, pFlags, (int)*rawIter, &param, &txtIter, pTxtCnt, pCsv, ppRecord, pRecordsCapa, pRecordsChunk, ppField, pFieldsCapa, pFieldsChunk);

  handle_end_of_data_(*pFlags, &txtIter, pCsv, *ppRecord, ppField, pFieldsCapa, *pTxtCnt);
}

static void parse_wide_csv_(void *txtIter, const wchar_t *rawIter, const size_t rawSize, int *const pFlags, const csv_param_t *const pParam, csv_t *const pCsv, csv_record_t **const ppRecord, size_t *const pRecordsCapa, size_t *const pRecordsChunk, csv_field_t **const ppField, size_t *const pFieldsCapa, size_t *const pFieldsChunk, size_t *const pTxtCnt)
{
  static const bool isNarrow = false;
  const csv_param_t param = (csv_param_t){ (int)(wchar_t)pParam->separator, (int)(wchar_t)pParam->commentPrefix, pParam->skipEmpty, pParam->fallbackNarrow };
  for (const wchar_t *const rawEnd = rawIter + rawSize / sizeof(wchar_t); rawIter != rawEnd; ++rawIter)
    handle_single_char_(isNarrow, pFlags, (int)*rawIter, &param, &txtIter, pTxtCnt, pCsv, ppRecord, pRecordsCapa, pRecordsChunk, ppField, pFieldsCapa, pFieldsChunk);

  handle_end_of_data_(*pFlags, &txtIter, pCsv, *ppRecord, ppField, pFieldsCapa, *pTxtCnt);
}

// ***** public functions: *****

csv_raw_t *read_csv_raw(const char *fileName)
{
  return read_csv_raw_(true, fileName);
}

#if defined(_WIN32) || defined(__CYGWIN__)
csv_raw_t *read_csv_raw_w(const wchar_t *fileName)
{
  return read_csv_raw_(false, fileName);
}
#endif

csv_raw_t *make_csv_raw(const void *bytes, size_t byteCount)
{
  if (!bytes || !byteCount)
    return NULL;

  csv_raw_t *pRaw = alloc_raw_(byteCount);
  memcpy(pRaw->data, bytes, byteCount);
  return pRaw;
}

csv_t *parse_csv(csv_raw_t *pRaw, const csv_param_t *pParam)
{
  if (!pRaw || !pRaw->dataSize || pParam->separator == '\0' || pParam->separator == '\n' || pParam->separator == '\r' || pParam->separator == '"' || pParam->commentPrefix == '\n' || pParam->commentPrefix == '\r' || pParam->commentPrefix == '"')
  {
    free(pRaw);
    return NULL;
  }

  const size_t bomLen = get_bom_len_(pRaw), rawSize = pRaw->dataSize - bomLen;
  if (!rawSize)
  {
    free(pRaw);
    return NULL;
  }

  csv_t *pCsv = (csv_t *)calloc(1, sizeof(csv_t));
  if (!pCsv)
    abort();

  const void *const rawBegin = pRaw->data + bomLen; // this is the position where reading begins (while writing back does always start at the `pRaw` pointer)
  int flags = 0;
  csv_record_t *pRecord = NULL;
  csv_field_t *pField = NULL;
  size_t recordsCapa = 0, recordsChunk = CSV_RECORDS_CHUNK_, fieldsCapa = 0, fieldsChunk = CSV_FIELDS_CHUNK_, txtCnt = 0;
  get_is_narrow_(pRaw, bomLen, pParam->fallbackNarrow) ?
    parse_narrow_csv_(pRaw, (const unsigned char *)rawBegin, rawSize, &flags, pParam, pCsv, &pRecord, &recordsCapa, &recordsChunk, &pField, &fieldsCapa, &fieldsChunk, &txtCnt) :
    parse_wide_csv_(pRaw, (const wchar_t *)rawBegin, rawSize, &flags, pParam, pCsv, &pRecord, &recordsCapa, &recordsChunk, &pField, &fieldsCapa, &fieldsChunk, &txtCnt);

  if (pCsv)
    create_sorted_header_(pCsv);

  return pCsv;
}

csv_field_t *get_by_index(const csv_t *pCsv, size_t recordIdx, size_t fieldIdx)
{
  return pCsv && recordIdx < pCsv->recordsCount && fieldIdx < pCsv->records[recordIdx].fieldsCount ? pCsv->records[recordIdx].fields + fieldIdx : NULL;
}

csv_field_t *get_by_name(const csv_t *pCsv, size_t recordIdx, const void *colName)
{
  if (!pCsv || recordIdx >= pCsv->recordsCount || !colName)
    return NULL;

  const csv_col_head_t find = { .colName = colName };
  const csv_col_head_t *const pFound = (csv_col_head_t *)bsearch(&find, pCsv->colHeads, pCsv->records->fieldsCount, sizeof(csv_col_head_t), pCsv->isNarrow ? comp_narrow_names_ : comp_wide_names_);
  return (pFound && pFound->colIdx < pCsv->records[recordIdx].fieldsCount) ? pCsv->records[recordIdx].fields + pFound->colIdx : NULL;
}

csv_field_t *repl_num_separators(csv_field_t *pNumField, int decSep, int thousandsSep, bool isNarrow)
{
  if (!pNumField)
    return pNumField;

  if (isNarrow)
  {
    const char dSep = (char)(unsigned char)decSep, tSep = (char)(unsigned char)thousandsSep;
    char *writeIter = pNumField->txt;
    for (const char *readIter = pNumField->txt; *readIter; ++readIter)
      if (*readIter != tSep)
        *writeIter++ = *readIter != dSep ? *readIter : '.';

    *writeIter = '\0';
    pNumField->txtLength = (size_t)(writeIter - pNumField->txt);
    return pNumField;
  }

  const wchar_t dSep = (wchar_t)decSep, tSep = (wchar_t)thousandsSep;
  wchar_t *writeIter = pNumField->wtxt;
  for (const wchar_t *readIter = pNumField->wtxt; *readIter; ++readIter)
    if (*readIter != tSep)
      *writeIter++ = *readIter != dSep ? *readIter : L'.';

  *writeIter = L'\0';
  pNumField->txtLength = (size_t)(writeIter - pNumField->wtxt);
  return pNumField;
}

void delete_csv(const csv_t *pCsv)
{
  if (!pCsv)
    return;

  free(pCsv->colHeads);
  free(pCsv->records->fields->txt); // this releases the pointer originally allocated for the `csv_raw_t` object
  for (csv_record_t *recIter = pCsv->records, *const end = recIter + pCsv->recordsCount; recIter != end; ++recIter)
    free(recIter->fields);

  free(pCsv->records);
  free((void *)(intptr_t)pCsv);
}

#if defined(__GNUC__)
#  pragma GCC diagnostic pop
#elif defined(_MSC_VER)
#  pragma warning(pop)
#endif
