/*
Copyright (c) 2022 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/// @file      ccsv.h
/// @brief     C interface to parse CSV content of a file.
/// @version   1.4
/// @author    Steffen Illhardt
/// @date      2022
/// @copyright MIT License
/// @pre       Requires compiler support for at least C11 (and platform-specific extensions).
/// @note      The program terminates abnormally in the event of a bad memory allocation.
/// @warning   Do not alter pointer values in the returned objects! They are used to release allocated memory. <br>
///            While string content of fields is mutable, updated strings must not exceed the original length.

#ifndef CCSV_H_INCLUDED__
#define CCSV_H_INCLUDED__

#ifndef __cplusplus
#  include <stddef.h>
#  include <stdbool.h>
#else
extern "C" {
#endif // __cplusplus
#include <stdlib.h>
#include <wchar.h>

#if defined(__GNUC__)
#  pragma GCC diagnostic push
#  pragma GCC diagnostic ignored "-Wpadded"
#elif defined(_MSC_VER)
#  pragma warning(push)
#  pragma warning(disable : 4820) // struct is padded
#endif

/// @brief Structure to store raw CSV byte data.
typedef struct
{
  size_t dataSize; ///< number of bytes
  unsigned char data[1]; ///< placeholder for the raw bytes of CSV data
} csv_raw_t;

/// @brief Structure specifying parameters for `parse_csv()`.
typedef struct
{
  int separator; ///< Character used to tokenize the data into fields. Depending on the character type of the raw data, this value must not exceed the size of a `char` or a `wchar_t`, respectively.
  int commentPrefix; ///< First character of a line, introducing a comment to be ignored. Depending on the character type of the raw data, this value must not exceed the size of a `char` or a `wchar_t`, respectively.
  bool skipEmpty; ///< Specify `true` to ignore empty lines.
  bool fallbackNarrow; ///< Hint which indicates the expected character type.
} csv_param_t;

/// @brief Structure to store the value of a field in CSV data. <br>
///        Use the `txt` member if `csv_t::isNarrow` is `true`. <br>
///        Use the `wtxt` member if `csv_t::isNarrow` is `false`.
typedef struct
{
  union
  {
    char *txt; ///< pointer to a zero-terminated string of `char` representing the value of a data field in **narrow** CSV data
    wchar_t *wtxt; ///< pointer to a zero-terminated string of `wchar_t` representing the value of a data field in **wide** CSV data
  };
  size_t txtLength; ///< number of characters in the data field, without terminating zero
} csv_field_t;

/// @brief Structure to store the array of fields in CSV data.
typedef struct
{
  csv_field_t *fields; ///< pointer to an array of data fields in a record
  size_t fieldsCount; ///< number of data fields in the array
} csv_record_t;

/// @brief Structure to store a column-header name along with its column index.
typedef struct
{
  const void *colName; ///< pointer to a zero-terminated string representing the name of a column-header
  size_t colIdx; ///< zero-based index of the column in the header record
} csv_col_head_t;

/// @brief Structure to store the array of records in CSV data, along with some meta data.
typedef struct
{
  csv_col_head_t *colHeads; ///< if applicable, pointer to a sorted list of column-headers
  csv_record_t *records; ///< pointer to an array of records of data fields
  size_t recordsCount; ///< number of records in the array of records
  size_t minFields; ///< lowest number of fields found in a record
  size_t maxFields; ///< highest number of fields found in a record
  bool isNarrow; ///< if `true` use `csv_field_t::txt`, if `false` use `csv_field_t::wtxt`
} csv_t;

#if defined(__GNUC__)
#  pragma GCC diagnostic pop
#elif defined(_MSC_VER)
#  pragma warning(pop)
#endif

/// @brief Read the raw byte data of a CSV file.
///
/// @param fileName  Name of the file as string of `char`.
///
/// @return Pointer to a `csv_raw_t` object. <br>
///         The returned pointer must be passed to exactly one `parse_csv()` call. <br>
///         NOTE: Do not reuse the returned pointer for subsequent `parse_csv()` calls, nor try to release it!
csv_raw_t *read_csv_raw(const char *fileName);

#if defined(_WIN32) || defined(__CYGWIN__) // Windows only
/// @brief Read the raw byte data of a CSV file. <br>
///        This function is Windows-only to support non-ASCII file names.
///
/// @param fileName  Name of the file as string of `wchar_t`.
///
/// @return Pointer to a `csv_raw_t` object. <br>
///         The returned pointer must be passed to `parse_csv()`.
csv_raw_t *read_csv_raw_w(const wchar_t *fileName);
#endif

/// @brief Prepare raw byte data of a string containing CSV.
///
/// @param bytes      Pointer to a string of `char` or `wchar_t`, containing the CSV data.
/// @param byteCount  Number of bytes in the string, without terminating zero.
///
/// @return Pointer to a `csv_raw_t` object. <br>
///         The returned pointer must be passed to exactly one `parse_csv()` call. <br>
///         NOTE: Do not reuse the returned pointer for subsequent `parse_csv()` calls, nor try to release it!
csv_raw_t *make_csv_raw(const void *bytes, size_t byteCount);

/// @brief Deserialize CSV data into a `csv_t` object.
///
/// @details <ul>
///   <li> The function determines whether the raw CSV data is based on `char` or `wchar_t`. <br>
///        A hint is passed which may indicate the expected character type. <br>
///        The `csv_t::isNarrow` member is set `true` in case of `char`-based data. </li>
///   <li> The Byte Order Mark of UTF-8 or UTF-16 LE / UTF-32 LE data is removed. </li>
///   <li> Records are separated by `\n`. All occurrences of `\r` are ignored (except in quoted strings). <br>
///        This way, the end of lines in the raw data can be `\r\n` as well as just `\n`. <br>
///        Each line represents a record of fields (unless a character introducing a comment is defined, <br>
///        which is an extension of this implementation). </li>
///   <li> The first line in the CSV data is not automatically assumed to be a header with column names. <br>
///        However, fields in named columns of records can be accessed using function `get_by_name()`. </li>
///   <li> The last line may or may not end with `\n`. </li>
///   <li> A defined separator is used for the tokenization of lines into fields. <br>
///        The separator can be any single character value, except of `\0`, `\r`, `\n`, or quotation mark. </li>
///   <li> An empty line generates a record with a single zero-length field <br>
///        (unless empty lines are skipped, which is an extension of this implementation). </li>
///   <li> A separator at the end of a line generates an additional zero-length field. </li>
///   <li> The function does not add zero-length fields to get all records aligned to the same number of fields. <br>
///        (Refer to the `csv_t::maxFields` member for the highest number of fields found in a record.) </li>
///   <li> Spaces and tabs in fields are preserved (unless defined as separator). </li>
///   <li> Quotation marks trigger a special processing: </li><ul>
///     <li> Quoted fields (or quoted substrings in fields, which is an extension of this implementation)
///          preserve separators, `\r`, and `\n` in the values. <br>
///          Examples: <br>
///          <PRE>  "foo\nbar"   -> foo\nbar </PRE>
///          <PRE>  foo"\r\n"bar -> foo\r\nbar </PRE></li>
///     <li> Two consecutive quotation marks within a quoted string are collapsed to a single quotation mark. <br>
///          Examples: <br>
///          <PRE>  "foo ""bar""" -> foo "bar" </PRE>
///          <PRE>  foo """bar""" -> foo "bar" </PRE>
///          <PRE>  """"          -> " </PRE></li>
///      <li> The surrounding quotation marks are always removed. </li></ul>
///   <li> Empty data will cause the function to return `NULL`. </li>
///   <li> Any `\0` character in the passed data invalidates the CSV and causes undefined behavior. </li></ul>
///
/// @param pRaw    Pointer returned by `read_csv_raw()`, `read_csv_raw_w()`, or `make_csv_raw()`. <br>
///                Ownership is taken over, do not reuse the pointer for subsequent `parse_csv()` calls. <br>
///                Passing `NULL` will cause the function to return `NULL`.
/// @param pParam  Pointer to a `csv_param_t` object, specifying the the settings for the CSV parser. <ul>
///                <li> `::separator` Character used to tokenize the data into fields. <br>
///                     Specifying `\0`, `\r`, `\n`, or quotation mark will cause the function to return `NULL`. </li>
///                <li> `::commentPrefix` First character of a line, introducing a comment to be ignored. <br>
///                     Specify `\0` to indicate that no character has this special meaning. <br>
///                     Specifying `\r`, `\n`, or quotation mark will cause the function to return `NULL`. </li>
///                <li> `::skipEmpty` Pass `true` to ignore empty lines. </li>
///                <li> `::fallbackNarrow` Hint which indicates the expected character type. <br>
///                     The function does always try to auto-detect the character type. And auto-detection will never
///                     fail on implementations where a `wchar_t` has a size of 4 bytes (*nix). However, very short data
///                     might be ambiguous where a `wchar_t` has a size of 2 bytes (Windows). In such an unlikely case,
///                     this parameter defines whether the function falls back to parse the data as `char`-based CSV
///                     (specify `true`), or as `wchar_t`-based CSV (specify `false`). <ul>
///                     <li> The value of this parameter will be ignored if a Byte Order Mark is found. </li>
///                     <li> Value `true` will be overridden if the raw data contains null-bytes (which indicate the presence of `wchar_t`). </li>
///                     <li> Value `false` will be overridden if the size of the raw data is not a multiple of the size of a `wchar_t`. </li></ul></li></ul>
///
/// @return Pointer to a `csv_t` object. <br>
///         If the `isNarrow` member is `true` use the `txt` member of the underlying `csv_field_t` objects. <br>
///         If the `isNarrow` member is `false` use the `wtxt` member of the underlying `csv_field_t` objects. <br>
///         Any returned non-NULL pointer must be passed to `delete_csv()` if not used anymore.
csv_t *parse_csv(csv_raw_t *pRaw, const csv_param_t *pParam);

/// @brief Access a field by its zero-based record and field indices.
///
/// @param pCsv       Pointer to the `csv_t` object.
/// @param recordIdx  Zero-based index of the record to be accessed.
/// @param fieldIdx   Zero-based index of the field to be accessed.
///
/// @return Pointer to the found `csv_field_t` object, or `NULL` if out of range. <br>
///         NOTE: Do not release the pointer with `free()`! The `csv_t` object still owns the data.
csv_field_t *get_by_index(const csv_t *pCsv, size_t recordIdx, size_t fieldIdx);

/// @brief Access a field by its zero-based record index and column-header name.
///
/// @param pCsv       Pointer to the `csv_t` object.
/// @param recordIdx  Zero-based index of the record to be accessed.
/// @param colName    Header name of the column in the record. (case-sensitive) <br>
///                   Pass a zero-terminated string of `char` if the `isNarrow` member of `pCsv` is `true`. <br>
///                   Pass a zero-terminated string of `wchar_t` if the `isNarrow` member of `pCsv` is `false`.
///
/// @return Pointer to the found `csv_field_t` object, or `NULL` if searching failed. <br>
///         NOTE: Do not release the pointer with `free()`! The `csv_t` object still owns the data.
csv_field_t *get_by_name(const csv_t *pCsv, size_t recordIdx, const void *colName);

/// @brief Replace decimal separator and thousands separator in a field to make a numeric string convertible
///        using the `strto...` functions or `wcsto...` functions, respectively.
///
/// @param pNumField     Pointer to the `csv_field_t` object containing the numeric string.
/// @param decSep        Decimal separator to be replaced.
/// @param thousandsSep  Thousands separator to be replaced. <br>
///                      Pass `\0` if no thousands separator is expected.
/// @param isNarrow      Value of the `isNarrow` member in the belonging `csv_t` object.
///
/// @return Pointer `pNumField`, pointing to the updated content.
csv_field_t *repl_num_separators(csv_field_t *pNumField, int decSep, int thousandsSep, bool isNarrow);

/// @brief Release allocated memory, returned by `parse_csv()`.
///
/// @param pCsv  Pointer to the `csv_t` object. `NULL` can be passed and will be ignored.
void delete_csv(const csv_t *pCsv);

#ifndef __cplusplus // C only
// clang-format off
#  define field_to_num(_type, _pNumField, _decSep, _thousandsSep, _isNarrow)                                               \
    (repl_num_separators(_pNumField, _decSep, _thousandsSep, _isNarrow),                                                   \
     _Generic((_type)0,                                                                                                    \
       long:               ((_isNarrow) ? strtol((_pNumField)->txt, NULL, 10)   : wcstol((_pNumField)->wtxt, NULL, 10)),   \
       long long:          ((_isNarrow) ? strtoll((_pNumField)->txt, NULL, 10)  : wcstoll((_pNumField)->wtxt, NULL, 10)),  \
       unsigned long:      ((_isNarrow) ? strtoul((_pNumField)->txt, NULL, 10)  : wcstoul((_pNumField)->wtxt, NULL, 10)),  \
       unsigned long long: ((_isNarrow) ? strtoull((_pNumField)->txt, NULL, 10) : wcstoull((_pNumField)->wtxt, NULL, 10)), \
       float:              ((_isNarrow) ? strtof((_pNumField)->txt, NULL)       : wcstof((_pNumField)->wtxt, NULL)),       \
       double:             ((_isNarrow) ? strtod((_pNumField)->txt, NULL)       : wcstod((_pNumField)->wtxt, NULL)),       \
       long double:        ((_isNarrow) ? strtold((_pNumField)->txt, NULL)      : wcstold((_pNumField)->wtxt, NULL))))
// clang-format on

#else // C++ only
} // extern "C" off

#  if defined(__GNUC__)
#    pragma GCC diagnostic push
#    pragma GCC diagnostic ignored "-Wunused-macros"
#    if defined(__clang__)
#      pragma GCC diagnostic ignored "-Wunneeded-internal-declaration"
#      pragma GCC diagnostic ignored "-Wc++98-compat-pedantic"
#      pragma GCC diagnostic ignored "-Wdocumentation-unknown-command"
#    endif
#  elif defined(_MSC_VER)
#    pragma warning(push)
#    pragma warning(disable : 4514) // unreferenced inline function has been removed
#  endif

/// @brief Convenience macro which replaces the separators in a field and converts the updated string to the specified numeric type. <br>
///        NOTE: The content of the `csv_field_t` object pointed to by `_pNumField` is getting altered.
///
/// @param _type          Name of the numeric type to which the conversion is performed. <br>
///                       Valid types are: `long`, `long long`, `unsigned long`, `unsigned long long`, `float`, `double`, and `long double`.
/// @param _pNumField     Pointer to the `csv_field_t` object containing the numeric string.
/// @param _decSep        Decimal separator (as `int`) to be replaced.
/// @param _thousandsSep  Thousands separator (as `int`) to be replaced. <br>
///                       Pass `\0` if no thousands separator is expected.
/// @param _isNarrow      Value (as `bool`) of the `isNarrow` member in the belonging `csv_t` object.
///
/// @return Output value of the conversion.
#  define field_to_num(_type, _pNumField, _decSep, _thousandsSep, _isNarrow) \
    ccsv_detail::field_to_num_impl_<_type>(repl_num_separators(_pNumField, _decSep, _thousandsSep, _isNarrow), _isNarrow)

/// @brief Callable object representing the deleter for C++ smart pointers of `csv_t`. <br>
///        It is used like so: <br>
///        `std::unique_ptr<const csv_t, csv_deleter_t> pCsv{ parse_csv(pRaw, &params), csv_deleter };`
#  if __cplusplus < 201703L
const // until C++17
#  else
constexpr inline // since C++17
#  endif
  auto csv_deleter{ [](const csv_t *csvPtr) noexcept { delete_csv(csvPtr); } };

/// @brief Type of the `csv_deleter` object.
using csv_deleter_t = decltype(csv_deleter);

namespace ccsv_detail // clang-format off
{
  template<typename T> T field_to_num_impl_(const csv_field_t *const pNumField, const bool isNarrow) noexcept = delete;

  template<> inline long field_to_num_impl_(const csv_field_t *const pNumField, const bool isNarrow) noexcept
  { return isNarrow ? strtol(pNumField->txt, nullptr, 10) : wcstol(pNumField->wtxt, nullptr, 10); }

  template<> inline long long field_to_num_impl_(const csv_field_t *const pNumField, const bool isNarrow) noexcept
  { return isNarrow ? strtoll(pNumField->txt, nullptr, 10) : wcstoll(pNumField->wtxt, nullptr, 10); }

  template<> inline unsigned long field_to_num_impl_(const csv_field_t *const pNumField, const bool isNarrow) noexcept
  { return isNarrow ? strtoul(pNumField->txt, nullptr, 10) : wcstoul(pNumField->wtxt, nullptr, 10); }

  template<> inline unsigned long long field_to_num_impl_(const csv_field_t *const pNumField, const bool isNarrow) noexcept
  { return isNarrow ? strtoull(pNumField->txt, nullptr, 10) : wcstoull(pNumField->wtxt, nullptr, 10); }

  template<> inline float field_to_num_impl_(const csv_field_t *const pNumField, const bool isNarrow) noexcept
  { return isNarrow ? strtof(pNumField->txt, nullptr) : wcstof(pNumField->wtxt, nullptr); }

  template<> inline double field_to_num_impl_(const csv_field_t *const pNumField, const bool isNarrow) noexcept
  { return isNarrow ? strtod(pNumField->txt, nullptr) : wcstod(pNumField->wtxt, nullptr); }

  template<> inline long double field_to_num_impl_(const csv_field_t *const pNumField, const bool isNarrow) noexcept
  { return isNarrow ? strtold(pNumField->txt, nullptr) : wcstold(pNumField->wtxt, nullptr); }
} // clang-format on

#  if defined(__GNUC__)
#    pragma GCC diagnostic pop
#  elif defined(_MSC_VER)
#    pragma warning(pop)
#  endif

#endif // __cplusplus
#endif // CCSV_H_INCLUDED__
