#include "ccsv.h"
#include <stdio.h>

// Outputting narrow and wide strings might be mutually exclusive depending on the implementation of stdout.
#define RUN_NARROW 1 // use ParseNCsv for CSV data consisting of char data (read from "test8.csv")
#define RUN_WIDE 1 // use ParseWCsv for CSV data consisting of wchar_t data (read from "test16.csv" on Windows, or from "test32.csv" on *nix)

int main(void)
{
  const csv_param_t params = {
    .separator = ';', // field separator
    .commentPrefix = '#', // first character in a line introducing a comment
    .skipEmpty = true, // skip empty lines
    .fallbackNarrow = true // fall back to parsing as `char`-based content
  };

#ifdef RUN_NARROW
  {
    puts("~~~~~~~ raw narrow");
    csv_raw_t *const pRaw = read_csv_raw("test8.csv");
    if (!pRaw)
      return 1;

    printf("%s\n", (char *)pRaw->data);

    puts("~~~~~~ narrow");
    const csv_t *pCsv = parse_csv(pRaw, &params);
    if (!pCsv)
      return 1;

    if (!pCsv->isNarrow)
    {
      puts("Whoops, this was supposed to be narrow data.");
      delete_csv(pCsv);
      return 1;
    }

    puts(" *** convert to double ***");
    printf("%f\n\n", field_to_num(double, // numeric type to convert the string to
                                  get_by_index(pCsv, 3, 0), // pointer to the field
                                  ',', // current decimal separator in the CSV data
                                  '.', // current thousands separator in the CSV data
                                  pCsv->isNarrow)); // indicate the character type of the string getting converted

    puts(" *** get value by named column ***");
    const csv_field_t *const pFound = get_by_name(pCsv, 2, "b");
    printf("  value of record at index 2 in column \"b\": |%s|\n", pFound ? pFound->txt : "*ERROR*");

    size_t line = 0;
    printf("\n(min. number of fields in a record: %zu)\n(max. number of fields in a record: %zu)\n", pCsv->minFields, pCsv->maxFields);
    puts(" *** using iterators ***");
    for (csv_record_t *recIter = pCsv->records, *const recordsEnd = recIter + pCsv->recordsCount; recIter != recordsEnd; ++recIter)
    {
      printf("%4zu:|", ++line);
      for (csv_field_t *fldIter = recIter->fields, *const fieldsEnd = fldIter + recIter->fieldsCount; fldIter != fieldsEnd; ++fldIter)
        printf("%s[%zu]|", fldIter->txt, fldIter->txtLength);

      puts("");
    }

    line = 0;
    puts(" *** using subscripts ***");
    for (size_t recIdx = 0, recStop = pCsv->recordsCount; recIdx != recStop; ++recIdx)
    {
      printf("%4zu:|", ++line);
      for (size_t fldIdx = 0, fldStop = pCsv->records[recIdx].fieldsCount; fldIdx != fldStop; ++fldIdx)
        printf("%s[%zu]|", pCsv->records[recIdx].fields[fldIdx].txt, pCsv->records[recIdx].fields[fldIdx].txtLength);

      puts("");
    }

    delete_csv(pCsv);

    static const char str[] = "123;456\n789;\"\"\";\"\"\"";
    pCsv = parse_csv(make_csv_raw(str, sizeof(str) - 1), &params);
    if (!pCsv)
      return 1;

    line = 0;
    puts("\n *** from string ***");
    for (size_t recIdx = 0, recStop = pCsv->recordsCount; recIdx != recStop; ++recIdx)
    {
      printf("%4zu:|", ++line);
      for (size_t fldIdx = 0, fldStop = pCsv->records[recIdx].fieldsCount; fldIdx != fldStop; ++fldIdx)
        printf("%s[%zu]|", pCsv->records[recIdx].fields[fldIdx].txt, pCsv->records[recIdx].fields[fldIdx].txtLength);

      puts("");
    }

    delete_csv(pCsv);

#  if defined(_WIN32) || defined(__CYGWIN__)
    csv_param_t paramsCpy = params;
    static const char ambiguous[] = "\x26\x21"; // on Windows, these bytes represent either two narrow characters '&' and '!', or one wide character "Ohm sign" U+2126
    puts("\n *** disambiguate bytes \\x26\\x21 using the `fallbackNarrow` parameter (only affects Windows) ***");

    paramsCpy.fallbackNarrow = true;
    pCsv = parse_csv(make_csv_raw(ambiguous, 2), &paramsCpy);
    fputs("  fallbackNarrow `true`, result: ", stdout);
    puts(pCsv->isNarrow ? "narrow" : "wide");
    delete_csv(pCsv);

    paramsCpy.fallbackNarrow = false;
    pCsv = parse_csv(make_csv_raw(ambiguous, 2), &paramsCpy);
    fputs("  fallbackNarrow `false`, result: ", stdout);
    puts(pCsv->isNarrow ? "narrow\n" : "wide\n");
    delete_csv(pCsv);
#  endif
  }
#endif

#if defined(RUN_WIDE) && (defined(_WIN32) || defined(__CYGWIN__)) // Windows (16 bit wchar_t)
  {
    _putws(L"\n~~~~~~ wide");
    const csv_t *pCsv = parse_csv(read_csv_raw_w(L"test16.csv"), &params);
    if (!pCsv)
      return 1;

    if (pCsv->isNarrow)
    {
      _putws(L"Whoops, this was supposed to be wide data.");
      delete_csv(pCsv);
      return 1;
    }

    _putws(L" *** convert to double ***");
    wprintf_s(L"%f\n\n", field_to_num(double, get_by_index(pCsv, 3, 0), ',', '.', pCsv->isNarrow));

    size_t line = 0;
    wprintf_s(L"(min. number of fields in a record: %zu)\n(max. number of fields in a record: %zu)\n", pCsv->minFields, pCsv->maxFields);
    for (csv_record_t *recIter = pCsv->records, *const recordsEnd = recIter + pCsv->recordsCount; recIter < recordsEnd; ++recIter)
    {
      wprintf_s(L"%4zu:|", ++line);
      for (csv_field_t *fldIter = recIter->fields, *const fieldsEnd = fldIter + recIter->fieldsCount; fldIter < fieldsEnd; ++fldIter)
        wprintf_s(L"%s[%zu]|", fldIter->wtxt, fldIter->txtLength);

      _putws(L"");
    }

    delete_csv(pCsv);

    static const wchar_t str[] = L"123;456\n789;\"\"\";\"\"\"";
    pCsv = parse_csv(make_csv_raw(str, sizeof(str) - sizeof(wchar_t)), &params);
    if (!pCsv)
      return 1;

    line = 0;
    _putws(L"\n *** from string ***");
    for (size_t recIdx = 0, recStop = pCsv->recordsCount; recIdx != recStop; ++recIdx)
    {
      wprintf_s(L"%4zu:|", ++line);
      for (size_t fldIdx = 0, fldStop = pCsv->records[recIdx].fieldsCount; fldIdx != fldStop; ++fldIdx)
        wprintf_s(L"%s[%zu]|", pCsv->records[recIdx].fields[fldIdx].wtxt, pCsv->records[recIdx].fields[fldIdx].txtLength);

      _putws(L"");
    }

    delete_csv(pCsv);
  }
#endif

#if defined(RUN_WIDE) && !defined(_WIN32) && !defined(__CYGWIN__) // POSIX (32 bit wchar_t)
  {
    fflush(stdout);
    stdout = freopen(NULL, "w", stdout); // try to re-open stdout to enable the output of wide strings even if narrow strings have been printed previously

    fputws(L"\n~~~~~~ wide\n", stdout);
    const csv_t *pCsv = parse_csv(read_csv_raw("test32.csv"), &params);
    if (!pCsv)
      return 1;

    if (pCsv->isNarrow)
    {
      fputws(L"Whoops, this was supposed to be wide data.\n", stdout);
      delete_csv(pCsv);
      return 1;
    }

    fputws(L" *** convert to double ***\n", stdout);
    wprintf(L"%f\n\n", field_to_num(double, get_by_index(pCsv, 3, 0), ',', '.', pCsv->isNarrow));

    size_t line = 0;
    wprintf(L"(min. number of fields in a record: %zu)\n(max. number of fields in a record: %zu)\n", pCsv->minFields, pCsv->maxFields);
    for (csv_record_t *recIter = pCsv->records, *const recordsEnd = recIter + pCsv->recordsCount; recIter < recordsEnd; ++recIter)
    {
      wprintf(L"%4zu:|", ++line);
      for (csv_field_t *fldIter = recIter->fields, *const fieldsEnd = fldIter + recIter->fieldsCount; fldIter < fieldsEnd; ++fldIter)
        wprintf(L"%S[%zu]|", fldIter->wtxt, fldIter->txtLength);

      putwchar(L'\n');
    }

    delete_csv(pCsv);

    static const wchar_t str[] = L"123;456\n789;\"\"\";\"\"\"";
    pCsv = parse_csv(make_csv_raw(str, sizeof(str) - sizeof(wchar_t)), &params);
    if (!pCsv)
      return 1;

    line = 0;
    fputws(L"\n *** from string ***\n", stdout);
    for (size_t recIdx = 0, recStop = pCsv->recordsCount; recIdx != recStop; ++recIdx)
    {
      wprintf(L"%4zu:|", ++line);
      for (size_t fldIdx = 0, fldStop = pCsv->records[recIdx].fieldsCount; fldIdx != fldStop; ++fldIdx)
        wprintf(L"%S[%zu]|", pCsv->records[recIdx].fields[fldIdx].wtxt, pCsv->records[recIdx].fields[fldIdx].txtLength);

      putwchar(L'\n');
    }

    delete_csv(pCsv);
  }
#endif

  return 0;
}
